#ifndef M6502_H
#define M6502_H

// See cpu.cc for copyrights.

#include "defs.h"
                               /* 6502 status flags:         */
#define	C_FLAG	  0x01         /* 1: Carry occured           */
#define	Z_FLAG	  0x02         /* 1: Result is zero          */
#define	I_FLAG	  0x04         /* 1: Interrupts disabled     */
#define	D_FLAG	  0x08         /* 1: Decimal mode            */
#define	B_FLAG	  0x10         /* Break [0 on stk after int] */
#define	R_FLAG	  0x20         /* Always 1                   */
#define	V_FLAG	  0x40         /* 1: Overflow occured        */
#define	N_FLAG	  0x80         /* 1: Result is negative      */

/** Rd6502()/Wr6502/Op6502() *********************************/
/** These functions are called when access to RAM occurs.   **/
/** They allow to control memory access. Op6502 is the same **/
/** as Rd6502, but used to read *opcodes* only, when many   **/
/** checks can be skipped to make it fast. It is only       **/
/** required if there is a #define FAST_RDOP.               **/
/************************************ TO BE WRITTEN BY USER **/
extern void Wr6502(register fWORD Addr, register fBYTE Value);
extern BYTE Rd6502(register fWORD Addr);
extern BYTE Op6502(register fWORD Addr) CONSTFUN;

extern BYTE Clocks[256];
extern BYTE ZNTable[256];
class m6502cpu
{
public:
    BYTE A,P,X,Y,S;
    PAIR PC;

private:
    fWORD ICount;
    bool IRQPending;
private:
    inline BYTE fetchop()
    {
    	register BYTE i = Op6502(PC.W++);
    	ICount += Clocks[i];
    	return i;
    }
    inline WORD M_LDWORD()
    {
		register WORD J = Op6502(PC.W++);
		J |= Op6502(PC.W++) << 8;
		return J;
    }
    
    /** Addressing Methods ***************************************/
    /** These macros calculate and return effective addresses.  **/
    /*************************************************************/
    WORD MC_Ab()	{ return M_LDWORD(); }
    WORD MC_Zp()	{ return BYTE(Op6502(PC.W++)); }
    WORD MC_Zx()	{ return BYTE(Op6502(PC.W++)+X); }
    WORD MC_Zy()	{ return BYTE(Op6502(PC.W++)+Y); }
    WORD MC_Ax()	{ return M_LDWORD()+X; }
    WORD MC_Ay()	{ return M_LDWORD()+Y; }
    WORD MC_Ix() { register WORD K = BYTE(Op6502(PC.W++) + X); return Op6502(K) | (Op6502(K+1) << 8); }
    WORD MC_Iy() { register WORD K = Op6502(PC.W++); return (Op6502(K) | (Op6502(K+1) << 8)) + Y; }
    /** Reading From Memory **************************************/
    /** These macros calculate address and read from it.        **/
    /*************************************************************/
    BYTE MR_Ab()	{ return Rd6502(MC_Ab()); }
    BYTE MR_Im()	{ return Op6502(PC.W++); }
    BYTE MR_Zp()	{ return Rd6502(MC_Zp()); }
    BYTE MR_Zx()	{ return Rd6502(MC_Zx()); }
    BYTE MR_Zy()	{ return Rd6502(MC_Zy()); }
    BYTE MR_Ax()	{ return Rd6502(MC_Ax()); }
    BYTE MR_Ay()	{ return Rd6502(MC_Ay()); }
    BYTE MR_Ix()	{ return Rd6502(MC_Ix()); }
    BYTE MR_Iy()	{ return Rd6502(MC_Iy()); }
    /** Writing To Memory ****************************************/
    /** These macros calculate address and write to it.         **/
    /*************************************************************/
    void MW_Ab(const BYTE r) { Wr6502(MC_Ab(), r); }
    void MW_Zp(const BYTE r) { Wr6502(MC_Zp(), r); }
    void MW_Zx(const BYTE r) { Wr6502(MC_Zx(), r); }
    void MW_Zy(const BYTE r) { Wr6502(MC_Zy(), r); }
    void MW_Ax(const BYTE r) { Wr6502(MC_Ax(), r); }
    void MW_Ay(const BYTE r) { Wr6502(MC_Ay(), r); }
    void MW_Ix(const BYTE r) { Wr6502(MC_Ix(), r); }
    void MW_Iy(const BYTE r) { Wr6502(MC_Iy(), r); }
    
    BYTE M_POP() { return Op6502(++S | 0x100); }
    void M_PUSH(const BYTE r) { Wr6502(0x100|S--, r); }
    void M_FL(const BYTE r) { P = (P & ~(Z_FLAG|N_FLAG)) | ZNTable[r]; }
    
    inline void M_CARRYUPD_HI(const BYTE r) { P = (P & ~C_FLAG) | (r >> 7);     }
    inline void M_CARRYUPD_LO(const BYTE r) { P = (P & ~C_FLAG) | (r & C_FLAG); }
    
    inline void M_LDA(const BYTE r) { M_FL(A = r); }    
    inline void M_LDX(const BYTE r) { M_FL(X = r); }    
    inline void M_LDY(const BYTE r) { M_FL(Y = r); }    
    inline void M_AND(const BYTE r) { M_FL(A &= r); }
    inline void M_ORA(const BYTE r) { M_FL(A |= r); }
    inline void M_EOR(const BYTE r) { M_FL(A ^= r); }
    inline void M_SBC(const BYTE r)
    {
        /* Warning! C_FLAG is inverted before SBC and after it */
    	/* D_FLAG handling has been deleted; NES does not use decimal mode. */
        register WORD J = A - r - (~P&C_FLAG);
        P = (P & ~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG)) | ZNTable[(BYTE)J];
        if((A^r) & (A ^ (BYTE)J) & 0x80) P |= V_FLAG;
        if(!(J & 0xFF00))P |= C_FLAG;
        A = (BYTE)J;
    }
    inline void M_ADC(const BYTE r)
    {
    	/* D_FLAG handling has been deleted; NES does not use decimal mode. */
        register WORD J = A + r + (P & C_FLAG);
        P = (P & ~(N_FLAG|V_FLAG|Z_FLAG|C_FLAG)) | ZNTable[(BYTE)J];
        if(~(A^r) & (A ^ (BYTE)J) & 0x80) P |= V_FLAG;
        if(J & 0xFF00)P |= C_FLAG;
        A = (BYTE)J;
    }
    inline void M_BIT(const BYTE r)
    {
    	P = (P & ~(N_FLAG|V_FLAG|Z_FLAG)) | (r&(N_FLAG|V_FLAG));
    	if(!(r & A))P |= Z_FLAG;
	}
	inline void M_CMP(const BYTE r2, const BYTE r1)
	{
        register WORD J = r1 - r2;
        P = (P & ~(N_FLAG|Z_FLAG|C_FLAG)) | ZNTable[(BYTE)J];
        if(!(J & 0xFF00))P |= C_FLAG;
	}
	inline void M_RTI()
	{
	    P = M_POP() | R_FLAG;
	    PC.B.l = M_POP();
        PC.B.h = M_POP();
    }
    inline void M_RTS()
    {
        PC.B.l = M_POP();
        PC.B.h = M_POP();
        ++PC.W;
    }
    inline WORD M_ABSJSR()
    {
		register WORD J = Op6502(PC.W);
		J |= Op6502(++PC.W) << 8;
		M_PUSH(PC.B.h);
		M_PUSH(PC.B.l);
		return J;
    }
    inline void M_BRK()
    {
        ++PC.W;
        M_PUSH(PC.B.h);
        M_PUSH(PC.B.l);
        M_PUSH(P | B_FLAG);
        P = (P|I_FLAG) & ~D_FLAG;
        PC.B.l = Rd6502(0xFFFE);
        PC.B.h = Rd6502(0xFFFF);
    }
    inline void M_PLP()
    {
        register BYTE i = M_POP();
        if(IRQPending && ((i^P)&~i&I_FLAG))
        {
            P = i | R_FLAG;
            Int(INT_IRQ);
        }
        else
            P = i | R_FLAG;
    }
    inline void M_INDIRJMP()
    {
        register PAIR K;
        K.W = M_LDWORD();
        PC.B.l = Rd6502(K.W);
        ++K.B.l;               // Note the 6502 bug!
        PC.B.h = Rd6502(K.W);  // Without the bug, this would be: ++K.W
    }
    inline void M_CLI() /* ENABLE interrupts */
    {
        if(IRQPending && (P & I_FLAG))
        {
            P &= ~I_FLAG;
            Int(INT_IRQ);
        }
        else
            P &= ~I_FLAG;
    }	
public:
	m6502cpu() : ICount(0), IRQPending(false)
	{
		// Do not reset here, as the memory
		// has probably not yet been initialized.
	}
    /** Reset6502() **********************************************/
    /** This function can be used to reset the registers before **/
    /** starting execution with Run6502(). It sets registers to **/
    /** their initial values.                                   **/
    /*************************************************************/
	void Reset()
	{
        A = X = Y = 0x00;
        P = Z_FLAG|R_FLAG;
        S = 0xFF;
        PC.B.l = Op6502(0xFFFC);
        PC.B.h = Op6502(0xFFFD);
        ICount = 0;
        IRQPending = false;
	}
	
    /** Int6502() ************************************************/
    /** This function will generate interrupt of a given type.  **/
    /** INT_NMI will cause a non-maskable interrupt. INT_IRQ    **/
    /** will cause a normal interrupt, unless I_FLAG set in R.  **/
    /*************************************************************/
    void Int(InterruptType Type);

    void Tick();
};

#endif
